/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <libintl.h>

#include "sambascanner.h"
#include "Mutex.h"
#include "Tools.h"
#include "String.h"
#include "IP_Range.h"
#include "Threads.h"
#include "Processes.h"
#include "ProcessTable.h"

extern struct Options_Structure Options;

ProcessTableID Exec_Timeout(const char *Command, const char *Parameter, double TermTime, double KillTime, int Outfile_FD) {
	pid_t PID;
	char **Arguments = NULL;
	int ArgumentCount = 0;
	ProcessTableID ProcID;
					
	int Modus[2];
	struct Redirect_Thread_InputData *Data;

	message(3, _("About to execute command: %s (%s) -> FD %d\n"), Command, Parameter, Outfile_FD);

	Data = malloc(sizeof(struct Redirect_Thread_InputData));
	if (pipe(Modus) == -1) {
		message(1, _("Error creating redirection pipe for execution of '%s' (%s): %s\n"), Command, Parameter, strerror(errno));
		exit(EXIT_FAILURE);
	}
	message(4, _("[pTID = %ld] Created pipe fd %d -> fd %d\n"), TID(pthread_self()), Modus[0], Modus[1]);
	
	Lock_Exec();
	ProcID = ProcessTable_CreatePreliminary(TermTime, KillTime);
	
	Data->Outfile_FD = Outfile_FD;
	Data->Read_FD = Modus[0];
	Data->ProcID = ProcID;
	Data->TermTime = TermTime;
	Data->KillTime = KillTime;
	Data->UniqueThreadID = GetUniqueThreadID();
	pthread_mutex_init(&Data->Active, NULL);
	pthread_mutex_lock(&Data->Active);
	{
		pthread_t ThreadID;
		message(4, _("Starting thread [UniqTID %d], Redirect_Thread (ReadFD = %d -> WriteFD = %d)\n"), Data->UniqueThreadID, Data->Read_FD, Data->Outfile_FD);
		if (pthread_create(&ThreadID, NULL, Redirect_Thread, Data)) {
			message(1, _("Error creating redirection thread for execution of '%s' (%s): %s\n"), Command, Parameter, strerror(errno));
			exit(EXIT_FAILURE);
		}
		ProcessTable_UpdateEntry_ThreadID(ThreadID);
		message(4, _("Thread [UniqTID %d] received pTID = %ld\n"), Data->UniqueThreadID, TID(ThreadID));
	}
	Unlock_Exec();
	
	PID = fork();
	if (PID == -1) {
		message(1, _("Fork failed: %s\n"), strerror(errno));
		exit(EXIT_FAILURE);
	}

	if (PID == 0) {
		/* Close reading end of pipe */
		close(Modus[0]);

		/* First argument is the caller */
		ArgumentCount++;
		Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
		Arguments[ArgumentCount-1] = (char*)malloc((strlen(Command)*sizeof(char))+1);
		strcpy(Arguments[ArgumentCount-1], Command);
	
		/* Parse the arguments */
		{
			char *Buffer;
			int Length;
			int i = 0;
			int Escaped = 0;

			Length = strlen(Parameter);
			Buffer = (char*)malloc(Length + 1);
			if (Buffer == NULL) {
				message(1, _("Malloc failed when allocating temporary buffer: %s\n"), strerror(errno));
				exit(EXIT_FAILURE);
			}
			
			char Tmp[2];
			Tmp[1] = 0;
			Buffer[0] = 0;
			while (i < Length) {
				if (Parameter[i] == '\'') {
					Escaped = !Escaped;
				} else if ((Parameter[i] == '\\') && (Parameter[i + 1] == '\'')) {
						Tmp[0] = '\'';
						strcat(Buffer, Tmp);	/* Add a >'< */
						i++;
				} else if ((!Escaped) && (Parameter[i] == ' ')) {
					/* Parameter was ended by space */
					ArgumentCount++;
					Arguments = (char**)realloc(Arguments, sizeof(char**) * ArgumentCount);
					Arguments[ArgumentCount - 1] = (char*)malloc((strlen(Buffer) + 1) * sizeof(char));
					strcpy(Arguments[ArgumentCount - 1], Buffer);
					Buffer[0] = 0;
				} else {
					/* Either a escaped space or a regular character */
					Tmp[0] = Parameter[i];
					strcat(Buffer, Tmp);
				}
				i++;
			}
			
			ArgumentCount++;
			Arguments = (char**)realloc(Arguments, sizeof(char**) * ArgumentCount);
			Arguments[ArgumentCount - 1] = (char*)malloc((strlen(Buffer) + 1) * sizeof(char));
			strcpy(Arguments[ArgumentCount - 1], Buffer);

			/* Add the NULL terminator argument */
			ArgumentCount++;
			Arguments = (char**)realloc(Arguments, sizeof(char**) * ArgumentCount);
			Arguments[ArgumentCount - 1] = NULL;
		}

		{
			int Old_Stdout, Old_Stderr;
			int DevNull;

			DevNull = open("/dev/null", O_WRONLY, 0644);
			if (DevNull == -1) {
				message(1, _("Opening /dev/null failed: %s\n"), strerror(errno));
				exit(EXIT_FAILURE);
			}
			
			errno = 0;
			Old_Stdout = dup(STDOUT_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup1: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			Old_Stderr = dup(STDERR_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup2: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			dup2(Modus[1], STDOUT_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup3: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
#ifdef DEBUG
			dup2(Modus[1], STDERR_FILENO);
#else
			dup2(DevNull, STDERR_FILENO);
#endif
			if (errno != 0) { fprintf(stderr, "Error dup4: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			execv(Command, Arguments);
			dup2(Old_Stdout, STDOUT_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup5: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			dup2(Old_Stderr, STDERR_FILENO);
			if (errno != 0) { fprintf(stderr, "Error dup6: %s\n", strerror(errno)); fflush(stdout); fflush(stderr); *(int*)NULL = 0; }
			message(1, _("Execution of '%s' failed: %s\n"), Command, strerror(errno));
		}
		exit(EXIT_FAILURE);
	} else {
		/* Close writing end of pipe */
		close(Modus[1]);
		
		/* Make entry in process list valid */
		ProcessTable_FinishEntry(PID);

		/* Parent process */
		message(3, _("Forked off child '%s %s' with PID %d.\n"), Command, Parameter, PID);
	}
	message(5, _("Entry of process (ProcID = %d) with PID %d done.\n"), ProcID, PID);
	message(4, _("Unlocking redirection to thread with [UniqTID %d]\n"), Data->UniqueThreadID);
	pthread_mutex_unlock(&Data->Active);	/* Start redirection */
	
	return ProcID;
}

int WaitFor(ProcessTableID ProcID) {
	double Before, After;
	int Return;
	int Status;
	
	Before = GetTime();
	do {
		message(5, _("Process sleeping until signalled: %d\n"), ProcID);
		Sleep_Until_Children_Dead();
	} while (ProcessTable_Get_Running(ProcID));
	After = GetTime();
	Accept_Dead_Child();
	message(5, _("[ProcID %d] has accepted dead child [PID %d] but waits for termination of Redirect_Thread\n"), ProcID, -1);

	/* Okay, the child died. The Redirect_Thread will soon exit. */
	Return = pthread_join(ProcessTable_Get_TID(ProcID), NULL);
	if (Return) {
		message(3, _("[ProcID %d] could not join thread: %s\n"), ProcID, strerror(Return));
		exit(EXIT_FAILURE);
	}
	message(5, _("[ProcID %d] has seen that its Redirect_Thread also exited.\n"), ProcID);

	Status = ProcessTable_Get_Status(ProcID);
	message(5, _("Process ProcID = %d left RUNNING state, Status = %d.\n"), ProcID, Status);

	Return = -1;
	if (WIFSIGNALED(Status)) {
		char *Msg = SignalMessage(Status);
		Return = -WTERMSIG(Status);
		if (Msg != NULL) {	
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, signalled by signal %s.\n"), ProcID, After - Before, Msg);
		} else {
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, signalled by signal number %d.\n"), ProcID, After - Before, -Return);
		}
	} else {
		if (WIFEXITED(Status)) {
			Return = WEXITSTATUS(Status);
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, exited with status %d.\n"), ProcID, After - Before, Return);
		} else {
			message(3, _("Waited for thread with ProcID = %d %.2f seconds, exited stange (status %x).\n"), ProcID, After - Before, Status);
		}
	}

	return Return;
}

